/*
 * MemCheck, a memory-checker for C/C++.
 * 
 * This file is released under the terms of the GNU General Public License
 * as they appear in http://www.gnu.org/licenses/gpl.txt.
 * 
 * Copyright 2001, 2002 erwin.rieger@ibrieger.de
 */

// test bcm (bounds checked malloc) 

#include <stdio.h>
#include <stdlib.h>
#include <bcm/bcm_rename.h>

#ifdef WIN32
  __declspec(dllimport)
#endif
extern int test_lib(int);

void errtests() {
  void *m;

  m=malloc(100);
  // overwrite bevore:
  ((char*)m)[-1]='0';
  // overwrite after:
  ((char*)m)[100]='0';
  free(m);

  // check malloc 0
  m=malloc(0);
  free(m);

  m=calloc(100, 1);
  free(m);

  // check calloc 0
  m=calloc(0, 1);
  free(m);

  m=malloc(1);
  m=realloc(m, 100);
  // overwrite bevore:
  ((char*)m)[-1]='0';
  // overwrite after:
  ((char*)m)[100]='0';
  m=realloc(m, 10);
  // overwrite bevore:
  ((char*)m)[-1]='0';
  // overwrite after:
  ((char*)m)[10]='0';
  free(m);
  // check free 0
  free(NULL);

  // call library bcmtestlib
  test_lib(0);
  }

void tests() {
  void *m;

  m=malloc(100);
  free(m);

  // check malloc 0
  m=malloc(0);
  free(m);

  m=calloc(100, 1);
  free(m);

  // check calloc 0
  m=calloc(0, 1);
  free(m);

  m=malloc(1);
  m=realloc(m, 100);
  m=realloc(m, 10);
  free(m);
  // check free 0
  free(NULL);
  // call library bcmtestlib
  test_lib(0);
  }

void myLogFunc(char *msg) {
  fprintf(stderr, "bcmtest.myLogFunc: %s", msg);
  }

int main(int argc, char **argv) {
  void *m;

  setLogFunction(myLogFunc);

  fprintf(stderr, "start main\n");
  // allocate mem without bcm:
  m=malloc(10);

  bcmInit(1);
  errtests();

  bcmInit(0);
  fprintf(stderr, "testing with bcm off:\n");
  tests();
  fprintf(stderr, "end testing with bcm off\n");

  bcmInit(1);
  // free mem with bcm, should generate 'unknown memory' error
  free(m);

  fprintf(stderr, "end main\n");
  exit(0);
  }





