#/*
# MemCheck, a memory-checker for C/C++.
# 
# This file is released under the terms of the GNU General Public License
# as they appear in http://www.gnu.org/licenses/gpl.txt.
# 
# Copyright 2001, 2002 erwin.rieger@ibrieger.de
#*/
#
# Funktionen fuer Makefile.in Generierung.
#


#
# Source mit dependency:
#   Trennung der einzelnen Sourcen durch ws.
#   cfile:dep1+dep2+dep3 cfile2:dep4+dep5 ...
# Shell Dict:
#   funktionen werden uebersichtlicher
#

c_suff=".c"
cpp_suff=".cc"
normal_lib_suffix=".a"
shared_lib_suffix=".so"
depend_suffix=".dep"
yacc_suffix=".y"
lex_suffix=".l"
idl_suffix=".idl"
depsep="+"               # Dependency separator, x:a.h+b.h+...


filestoclean=""
alltargets=""
saved_ifs=""
dict_val=""
buildsubdir=""
deplog="./missing.depfiles"
echo "" > $deplog

function get_deplog {
  echo "$deplog"
  }

#
# Extrahiert value aus Shell-Dict
# Parameter: dict, keyname
#
function get_dict_val {
  # dict_val=""
  saved_ifs="$IFS"
  IFS=","
  for part in $1; do
    # restore IFS:
    IFS="$saved_ifs"
    key=`echo $part|cut -d':' -f1`
    if test "$key" = "$part"; then
      echo "invalid shell dict in get_dict_val(): $1" 1>&2
      exit -200
      fi
    # echo "testing key: $key"
    if test "$key" = "$2"; then
      # echo "found $key" 1>&2
      echo $part|cut -d':' -f2-
      fi
    done
  # restore IFS:
  IFS="$saved_ifs"
  }

#
# Ersetzt Wert in dict durch einfaches anfuegen am Anfang,
# ist abhaengig vom Algorithmus in get_dict_val.
#   Parameter: dict, key and val
function dict_insert {
  echo "$2,$1"
  }

##########################################################################
#                                                                        #
# Function CREATE_MAKEFILE_IN                                            #
#                                                                        #
##########################################################################
#
# Generiert Makefile.in Header
#   Parameter:  build -  subdir
#
function create_makefile_in {
  buildsubdir="$1"
  echo ""
  echo "# Open Shop Makefile Template in $./$1"
  echo ""
  echo "# Common to all makefiles:"
  echo ""
  echo "#"
  echo "# Anmerkungen zu den Make Variablen:"
  echo "#"
  echo "#   CFLAGS werden zum compilieren von C Sourcen"
  echo "#   verwendet. Fuer C++ Sourcen sind CXXFLAGS zustaendig, "
  echo "#   sie werden analog zu CFLAGS gebildet und werden desshalb"
  echo "#   hier nicht beschrieben. SHCFLAGS bzw. SHCXXFLAGS sind"
  echo "#   die entspr. Variablen plus Optionen um shared binaries"
  echo "#   zu erzeugen (-fpic -shared)."
  echo "#"
  echo "#   CFLAGS wird aus ADDCFLAGS, Conf - CFLAGS, INCLUDES DEFS und "
  echo "#   konstanten, globalen Optionen, z.B. -Wall gebildet."
  echo "#"
  echo "#   ADDCFLAGS sind die von externen Modulen benoetigten Flags"
  echo "#   (--with-xxx) z.B. mysql oder memcheck"
  echo "#   Conf - CFLAGS sind die von 'configure' gelieferten, system-  "
  echo "#   abhaengigen Flags."
  echo "#"
  echo "#   INCLUDES sind alle noetigen -I Optionen, siehe unten."
  echo "#"
  echo "#   DEFS sind die von 'configure' gelieferten Defines, z.B."
  echo "#   -DYYTEXT_POINTER=1."
  echo "#"
  echo "#   INCLUDES wird aus ADDINCLUDES, globalen -I Optionen wie "
  echo "#   -I SRCDIR/include und LOCALINCLUDE gebildet."
  echo "#   "
  echo "#   ADDINCLUDES sind wieder von extenen Modulen benoetigte"
  echo "#   Include Pfade."
  echo "#   "
  echo "#   LOCALINCLUDE dient dazu, den Include - Pfad fuer einzelne Makefiles"
  echo "#   zu modifizieren (z.B. werden so die CORBA Includes behandelt)."
  echo "#"
  echo ""
  echo "# Programs: "
  echo "         CC=@CC@"
  echo "        CXX=@CXX@"
  echo "    INSTALL=@INSTALL@"
  echo "       YACC=@YACC@"
  echo "        LEX=@LEX@"
  echo "        IDL=idl"
  echo "         MV=mv"
  echo ""
  echo "# Directories: "
  echo "     SRCDIR=@srcdir@"
  echo "     PREFIX=@prefix@"
  echo "  TOPSRCDIR=@top_srcdir@"
  echo "   BUILDDIR=@builddir@"
  echo "   MICO_DIR=@MICO_DIR@"
  echo "   PZWO_DIR=@PZWO_DIR@"
  echo ""
  echo "# Flags: "
  echo "  SHLDFLAGS=-rdynamic"
  echo "ADDINCLUDES=@ADDINCLUDES@"
  echo "   INCLUDES=\$(ADDINCLUDES) -I./. -I\$(TOPSRCDIR)/include \\
                    \$(LOCALINCLUDE) "
  echo "       DEFS=@DEFS@"
  echo "  ADDCFLAGS=@ADDCFLAGS@"
  echo "     CFLAGS=\$(ADDCFLAGS) @CFLAGS@ \$(INCLUDES) -Wall \\
                    \$(DEFS)"
  echo "   CXXFLAGS=\$(ADDCFLAGS) @CXXFLAGS@ \$(INCLUDES) -Wall \\
                    \$(DEFS)"
  echo "   SHCFLAGS=\$(CFLAGS) -fpic -shared"
  echo " SHCXXFLAGS=\$(CXXFLAGS) -fpic -shared"
  echo "   CPPFLAGS=@CPPFLAGS@"
  echo "    LDFLAGS=@LDFLAGS@"
  echo "  YACCFLAGS=-d"
  echo "     LEXLIB=@LEXLIB@"
  echo ""
  echo "# System - Libs: "
  echo "       LIBS=@LIBS@"
  echo "    ADDLIBS=@ADDLIBS@"
  echo "  PGSYSLIBS=\$(ADDLIBS) -L/usr/lib \$(LIBS)"
  echo "PREPSYSLIBS=\$(ADDLIBS) -L/usr/lib \$(LIBS) @LEXLIB@"
  echo "   MICOLIBS=-L\$(MICO_DIR)/lib -lmico2.2.4 -lmicocoss2.2.4 -lstdc++"
  echo ""
  echo "# End of common section"

  filestoclean=""
  alltargets=""
  }

#
#   Parameter:
#     1 src
#
function get_src { 
  if test -n "`echo $1|grep ':'`"; then
    echo $1|cut -d':' -f1
  else
    echo $1
    fi
  }

#
# Trennt dependencies von der Source (a:x.h+b.h+...)
#
function get_deps_from_src {
  if test -n "`echo $1|grep ':'`"; then
    other_dep="`echo $1|cut -d':' -f2-`"
    # echo "other : $other_dep" >&2
    saved_ifs="$IFS"
    IFS="$depsep"
    for dep in $other_dep; do
      # echo "single dep : $dep" >&2
      echo "$dep"
      done
    # restore IFS:
    IFS="$saved_ifs"
    fi
  }

#
# Liest dependency aus src (durch $depsep getrennt) und aus
# dependency-file aus:
#   Parameter:
#     1 src
#
function get_dep { 
  # deps in source:
  get_deps_from_src "$1"

  src="`get_src $1`"
  dep="./$buildsubdir/$src$c_suff$depend_suffix"
  # echo "depfile: $dep" >&2
  if test -f "$dep"; then
    # echo "depfile: `cat $dep`" >&2
    cat $dep
  else
    echo "missing dependency-file: $dep" >> $deplog
    fi
  }

#
# Liest dependency aus src (durch $depsep getrennt) und aus
# dependency file aus:
#   Parameter:
#     1 src
#
function get_dep_cpp {
  get_deps_from_src "$1"

  src="`get_src $1`"
  dep="$buildsubdir/$src$cpp_suff$depend_suffix"
  # echo "depfile: $dep" >&2
  if test -f "$dep"; then
    # echo "depfile: `cat $dep`" >&2
    cat $dep
  else
    echo "missing dependency-file: $dep" >> $deplog
    fi
  }

##########################################################################
#                                                                        #
# Function MAKEDEPEND_RULE                                               #
#                                                                        #
##########################################################################
#
# Generiert depend rule
#   Parameter: dict mit keys:
#     csrcs
#     cxxsrcs
#     shared
#     othersrcdir
#
function makedepend_rule {
  _csrcs=`get_dict_val "$1" "csrcs"`
  _cxxsrcs=`get_dict_val "$1" "cxxsrcs"`
  _shared=`get_dict_val "$1" "shared"`
  echo ""
  echo "# depend rules:"
  alldeps=""

  for name in $_csrcs; do
    src="`get_src $name`$c_suff"
    # dep="$src$depend_suffix"
    dep="./$src$depend_suffix"
    alldeps="$alldeps $dep"
    echo ""
    src="\$(SRCDIR)/$src"
    echo "$dep: $src"
    if test -n "$_shared"; then
      echo "	makedepend -- \$(SHCFLAGS) -- -w10000 -a  -Y -f- $src 2>/dev/null|grep ':'|cut -d':' -f2 > $dep"
    else
      echo "	makedepend -- \$(CFLAGS) -- -w10000 -a  -Y -f- $src 2>/dev/null|grep ':'|cut -d':' -f2 > $dep"
      fi
    done

  for name in $_cxxsrcs; do
    src="`get_src $name`$cpp_suff"
    dep="./$src$depend_suffix"
    alldeps="$alldeps $dep"
    echo ""
    src="\$(SRCDIR)/$src"
    echo "$dep: $src"
    if test -n "$_shared"; then
      echo "	makedepend -- \$(SHCFLAGS) -- -w10000 -a  -Y -f- $src 2>/dev/null|grep ':'|cut -d':' -f2 > $dep"
    else
      echo "	makedepend -- \$(CFLAGS) -- -w10000 -a  -Y -f- $src 2>/dev/null|grep ':'|cut -d':' -f2 > $dep"
      fi
    done

  echo "depend: $alldeps"
  echo ""
  echo "depclean:"
  echo "	rm -f $alldeps"
  echo ""
  echo "# depend rules end"
  }

##########################################################################
#                                                                        #
# Function YACC_RULE                                                     #
#                                                                        #
##########################################################################
#
# Generiert yacc call:
#   Parameter ist ein dict mit den keys:
#     yaccfile
#     shared
#     
function yacc_rule {
  _yaccfile=`get_dict_val "$1" "yaccfile"`
  _shared=`get_dict_val "$1" "shared"`

  echo ""
  echo "# yaccing $_yaccfile$yacc_suffix:"
  echo ""
  in="\$(SRCDIR)/$_yaccfile"
  echo "$_yaccfile$c_suff: $_yaccfile.h"
  echo "$_yaccfile.h: $in$yacc_suffix"
  echo "	\$(YACC) \$(YACCFLAGS) $in$yacc_suffix"
  echo "	\$(MV) y.tab.c $_yaccfile$c_suff"
  echo "	\$(MV) y.tab.h $_yaccfile.h"
  echo ""

  comp_params=`dict_insert "$1" "csrcs:$_yaccfile:$_yaccfile$c_suff+$_yaccfile.h, othersrcdir:."`
  normal_compile_rule "$comp_params"

  echo "# yacc rule end"

  add_clean "filestoclean:$_yaccfile.h $_yaccfile$c_suff $_yaccfile.o"
  }

##########################################################################
#                                                                        #
# Function IDL_RULE                                                      #
#                                                                        #
##########################################################################
#
# Generiert idl call:
#   Parameter: dict mit den keys:
#     idlfile
#     shared
#     idlopts         # Additional options for idl compiler (-I...)
#     newname         # if given: rename idl outputfiles to newname[.cc|.h]
#     
function idl_rule {
  _idlfile=`get_dict_val "$1" "idlfile"`
  idlfile="$_idlfile$idl_suffix"
  _idlopts=`get_dict_val "$1" "idlopts"`
  _newname=`get_dict_val "$1" "newname"`

  in="\$(SRCDIR)/$idlfile"
  if test -n "$_newname"; then
    out="$_newname"
  else
    out="$_idlfile"
    fi
  echo ""
  echo "# idl $idlfile:"
  echo ""
  echo "$out$cpp_suff: $out.h"
  echo "$out.h: $in"
  echo "	\$(IDL) $_idlopts $in"
  if test -n "$_newname"; then
  echo "	mv $_idlfile$cpp_suff $out$cpp_suff"
  echo "	mv $_idlfile.h $out.h"
    fi
  echo ""

  newdict=`dict_insert "$1" "cxxsrcs:$out:$out$cpp_suff+$out.h,
                             othersrcdir:."`
               
  normal_compile_rule "$newdict"

  echo "# idl rule end"

  add_clean "filestoclean:$out.h $out$cpp_suff $out.o"
  }

##########################################################################
#                                                                        #
# Function INSTALL_RULE                                                  #
#                                                                        #
##########################################################################
#
# Install target:
#   Parameter: dict mit keys:
#     file
#     destdir
#     renameto     # file als destdir/renameto installieren
#
function install_rule {
  _file=`get_dict_val "$1" "file"`
  _destdir=`get_dict_val "$1" "destdir"`
  _renameto=`get_dict_val "$1" "renameto"`
  echo ""
  echo "# Install $_file:"
  echo "install: $_file"
  echo "	\$(INSTALL) -d $_destdir # Verzeichnis anlegen"
  if test -n "$_renameto"; then
    echo "	\$(INSTALL) $_file $_destdir/$_renameto"
  else
    echo "	\$(INSTALL) $_file $_destdir"
    fi
  echo "# End install"
  }

##########################################################################
#                                                                        #
# Function LEX_RULE                                                      #
#                                                                        #
##########################################################################
#     
# Generiert lex call:
#   Parameter: dict mit keys:
#     lexfile
#     shared
#     otherdeps
#     
function lex_rule {
  _lexfile=`get_dict_val "$1" "lexfile"`
  _otherdeps=`get_dict_val "$1" "otherdeps"`
  echo ""
  echo "# lexing $_lexfile$lex_suffix:"
  echo ""
  in="\$(SRCDIR)/$_lexfile"
  echo "$_lexfile$c_suff: $in$lex_suffix $_otherdeps"
  echo "	\$(LEX) $in$lex_suffix"
  echo "	\$(MV) lex.yy.c $_lexfile$c_suff"
  echo ""

  comp_params=`dict_insert "$1" "csrcs:$_lexfile:$_lexfile$c_suff,
                                 othersrcdir:."`
  normal_compile_rule "$comp_params"

  add_clean "filestoclean:$_lexfile$c_suff $_lexfile.o"

  echo "# lex rule end"
  }

##########################################################################
#                                                                        #
# Function NORMAL_COMPILE_RULE                                           #
#                                                                        #
##########################################################################
#
# Generiert Rule zum Uebersetzen einer Menge von C/C++ Sourcen,
# srcs werden zu den deps hinzugefuegt
#   Parameter: dict mit folgenden Keys:
#       csrcs, optional
#       cxxsrcs, optional
#       shared, optional
#       othersrcdir, optional
#       othercflags, optional
#
function normal_compile_rule {
  _csrcs=`get_dict_val "$1" "csrcs"`
  _cxxsrcs=`get_dict_val "$1" "cxxsrcs"`
  _shared=`get_dict_val "$1" "shared"`
  _othersrcdir=`get_dict_val "$1" "othersrcdir"`
  _othercflags=`get_dict_val "$1" "othercflags"`

  echo ""
  echo "# Compile objects:"

  if test -n "$_othersrcdir"; then
    source_dir="$_othersrcdir"
  else
    source_dir="\$(SRCDIR)"
    fi
  if test -n "$_othercflags"; then
    eff_cflags="$_othercflags"
    eff_cxxflags="$_othercflags"
  else
    if test -n "$_shared"; then
      eff_cflags="\$(SHCFLAGS)"
      eff_cxxflags="\$(SHCXXFLAGS)"
    else
      eff_cflags="\$(CFLAGS)"
      eff_cxxflags="\$(CXXFLAGS)"
      fi
    fi

  for cname in $_csrcs; do
    echo ""
    src=`get_src $cname`
    deps=`get_dep "$cname"`
    echo "$src.o: $source_dir/$src$c_suff"
    for dep in $deps; do
      echo "$src.o: $dep"
      done
    echo "	\$(CC) $eff_cflags -c $source_dir/$src$c_suff"
    done

  for cname in $_cxxsrcs; do
    echo ""
    # get_src "$cname"
    src=`get_src $cname`
    deps=`get_dep_cpp "$cname"`
    echo "$src.o: $source_dir/$src$cpp_suff"
    for dep in $deps; do
      echo "$src.o: $dep"
      done
    echo "	\$(CXX) $eff_cxxflags -c $source_dir/$src$cpp_suff"
    done

  echo ""
  echo "# Compile objects end"
  }

##########################################################################
#                                                                        #
# Function NORMAL_LIB_RULE                                               #
#                                                                        #
##########################################################################
#
# Generiert Rule fuer normale library
#   Parameter dict mit folgenden keys:
#     libname    # ohne 'lib' und ohne suffix
#     deps       # Dependencies dieser lib, srcs werden zu den 
#                # deps hinzugefuegt
#     srcs       # c Sources
#     cxxsrcs    # c++ Sources
#     libs       # libs werden zu den deps hinzugefuegt, in builddir/lib
#     sharedlibs # sharedlibs werden zu den deps hinzugefuegt, in builddir/lib
#     otherlibs  # Systemlibs, keine dependency
#     shared
#     install    # install lib in builddir/lib
#
function normal_lib_rule {
  _libname=`get_dict_val "$1" "libname"`
  _deps=`get_dict_val "$1" "deps"`
  _csrcs=`get_dict_val "$1" "csrcs"`
  _cxxsrcs=`get_dict_val "$1" "cxxsrcs"`
  _libs=`get_dict_val "$1" "libs"`
  _sharedlibs=`get_dict_val "$1" "sharedlibs"`
  _shared=`get_dict_val "$1" "shared"`
  _install=`get_dict_val "$1" "install"`
  _otherlibs=`get_dict_val "$1" "otherlibs"`

  if test -n "$_shared"; then
    libname="lib$_libname$shared_lib_suffix"
  else
    libname="lib$_libname$normal_lib_suffix"
    fi

  echo ""
  echo "# Compile library $libname:"
  echo ""
  if test -n "$_install"; then
    destlib="\$(BUILDDIR)/lib/$libname"
    add_all_target "targets:$destlib"
    echo "$destlib: $libname"
  echo "	\$(INSTALL) -d \$(BUILDDIR)/lib # Verzeichnis anlegen"
    echo "	\$(INSTALL) $libname \$(BUILDDIR)/lib"
    echo ""
  else
    add_all_target "targets:$libname"
    fi

  normal_compile_rule "$1"

  objs=""
  alllibs=""
  for cname in $_csrcs; do
    # get_src "$cname"
    objs="$objs `get_src $cname`.o"
    done
  for cname in $_cxxsrcs; do
    # get_src "$cname"
    objs="$objs `get_src $cname`.o"
    done
  for lname in $_libs; do 
    alllibs="$alllibs \$(BUILDDIR)/lib/lib$lname$normal_lib_suffix"
    # echo "allibs a: $alllibs" >&2
    done
  for lname in $_sharedlibs; do 
    alllibs="$alllibs \$(BUILDDIR)/lib/lib$lname$shared_lib_suffix"
    # echo "allibs so: $alllibs" >&2
    done

  add_clean "filestoclean:lib$_libname$shared_lib_suffix lib$_libname$normal_lib_suffix $objs"

  echo "Linking $libname:"
  echo "$libname: $objs $alllibs $_deps"
  if test -n "$_shared"; then
    # -rdynamic -shared
    echo "	\$(CC) \$(SHLDFLAGS) -shared -o $libname $objs $alllibs $_otherlibs"
  else
    # echo "	\$(CC) -o $libname $objs $alllibs"
    echo "	\$(AR) -qcs $libname $objs $alllibs $_otherlibs"
    fi
  echo ""
  echo "# Compile lib $libname end"
  }

function get_libs {
  for lname in $1; do 
    if test -n "$2"; then
      echo -n " \$(BUILDDIR)/lib/lib$lname$shared_lib_suffix"
    else
      echo -n " \$(BUILDDIR)/lib/lib$lname$normal_lib_suffix"
      fi
    done
  }

function get_libs2 {
  echo -n "-L\$(BUILDDIR)/lib "
  for lname in $1; do
    echo -n " -l$lname"
    done
  }

##########################################################################
#                                                                        #
# Function NORMAL_PROG_RULE                                              #
#                                                                        #
##########################################################################
#
# Generiert Rule fuer normales Programm
#   Parameter: dict mit keys:
#     progname
#     otherdeps      # srcs werden zu den deps hinzugefuegt
#     srcs
#     cxxsrcs
#     libs           # libs werden zu den deps hinzugefuegt
#     sharedlibs     # libs werden zu den deps hinzugefuegt
#     shared
#     systemlibs
#     libstring      # Alle libs, dadurch kann die Reihenfolge beim
#                    # linken festgelegt werden
#
function normal_prog_rule {
  _progname=`get_dict_val "$1" "progname"`
  _otherdeps=`get_dict_val "$1" "otherdeps"`
  _csrcs=`get_dict_val "$1" "csrcs"`
  _cxxsrcs=`get_dict_val "$1" "cxxsrcs"`
  _libs=`get_dict_val "$1" "libs"`
  _sharedlibs=`get_dict_val "$1" "sharedlibs"`
  _shared=`get_dict_val "$1" "shared"`
  _systemlibs=`get_dict_val "$1" "systemlibs"`
  _libstring=`get_dict_val "$1" "libstring"`
  echo ""
  echo "# Compile program $_progname:"

  add_all_target "targets:$_progname"
  normal_compile_rule "$1"

  libstr=`get_libs2 "$_libstring"`

  objs=""
  for cname in $_csrcs; do
    # get_src "$cname"
    objs="$objs `get_src $cname`.o"
    done
  for cname in $_cxxsrcs; do
    # get_src "$cname"
    objs="$objs `get_src $cname`.o"
    done
  add_clean "filestoclean:$_progname $objs"

  libs=`get_libs "$_libs" ""`
  shlibs=`get_libs "$_sharedlibs" "j"`
  echo "$_progname: $objs $libs $shlibs $_otherdeps"

  if test -n "$_shared"; then
    echo "	\$(CC) -o $_progname \$(SHLDFLAGS) $objs $libstr $_systemlibs"
  else
    echo "	\$(CC) -o $_progname \$(LDFLAGS) $objs $libstr $_systemlibs"
    fi
  echo ""
  echo "# Compile program $_progname end"
  }

##########################################################################
#                                                                        #
# Function CLEAN_RULE                                                    #
#                                                                        #
##########################################################################
#
# Generiert clean rule, es werden zusaetzlich alle durch add_clean
# hinzugefuegten files geloescht.
#
function clean_rule {
  echo ""
  echo "# clean rule:"
  echo "clean:"
  echo "	rm -f $filestoclean"
  echo ""
  }

##########################################################################
#                                                                        #
# Function ADD_CLEAN                                                     #
#                                                                        #
##########################################################################
#
#   Parameter: dict mit keys:
#     filestoclean
#
function add_clean {
  _filestoclean=`get_dict_val "$1" "filestoclean"`
  filestoclean="$filestoclean $_filestoclean"
  }


##########################################################################
#                                                                        #
# Function ALL_TARGET                                                    #
#                                                                        #
##########################################################################
#
# Generiert all target 
#
function all_target {
  echo ""
  echo "# all target:"
  echo "all: $alltargets"
  echo ""
  }

##########################################################################
#                                                                        #
# Function ADD_ALL_TARGET                                                #
#                                                                        #
##########################################################################
#
#   Parameter: dict mit keys:
#     targets
#
function add_all_target {
  _targets=`get_dict_val "$1" "targets"`
  alltargets="$alltargets $_targets"
  }

##########################################################################
#                                                                        #
# Function SUBDIR_RULE                                                   #
#                                                                        #
##########################################################################
#
# Generiert target rule fuer subdirs
#   Parameter: dict mit keys:
#     subdirs
#     target
#
function subdir_rule {
  _subdirs=`get_dict_val "$1" "subdirs"`
  _target=`get_dict_val "$1" "target"`
  echo ""
  echo "# make $_target in subdirs:"
  echo "$_target:"
  for dn in $_subdirs; do
    echo -n "	"
    echo "cd $dn; make $_target; cd .."
    done
  echo "# end make $_target"
  }

