/*
 * MemCheck, a memory-checker for C/C++.
 * 
 * This file is released under the terms of the GNU General Public License
 * as they appear in http://www.gnu.org/licenses/gpl.txt.
 * 
 * Copyright 2001, 2002 erwin.rieger@ibrieger.de
 */
/*
 *
 */

#ifndef __bcm_h__
#define __bcm_h__

#if ! defined(WIN32)
  #include <malloc.h>
#endif

#define BCMDUMMY "1v2v3v4v5v6v8m9m8h7h6h5h4h3h2h1" /* 32 Bytes, must be word aligned */
#define BCMMAGIC "**bcm**" // 8 bytes, must be word aligned
// #define FILL_CHAR ((char)'*')    /* Nicht 0, da dadurch Strings term. werden */
#define MemFromUsrMem(p) ((void*)((char*)p - sizeof(bcmalloc)))
#define UsrMemFromMem(p) ((void*)((char*)p + sizeof(bcmalloc)))
#define BCMUNKNOWN "unknown"
#if !defined(BCMOUT)
  #define BCMOUT stderr
#endif

#ifdef WIN32
  // bcm now static lib
  #ifdef BCMMAIN_EXPORTS
    #define BCMMAIN_API __declspec(dllexport)
  #else
    #define BCMMAIN_API __declspec(dllimport)
  #endif

  BCMMAIN_API void * (*realmallocptr) (size_t);
  BCMMAIN_API void (*realfreeptr) (void *);
  BCMMAIN_API void * (*realreallocptr) (void*, size_t);
  BCMMAIN_API void * (*mallocptr) (size_t, char *, int);
  BCMMAIN_API void * (*reallocptr) (void *, size_t, char *, int);
  BCMMAIN_API void (*freeptr) (void *, char *, int);
  BCMMAIN_API void (*logptr) (char *);
#else
  #define BCMMAIN_API /* notused */
#endif

typedef struct {
  char magic[sizeof(BCMMAGIC)]; // 8 bytes
  char *file;                   // 4
  int  line;                    // 4
  unsigned int  size;           // 4
  char dummy[sizeof(BCMDUMMY)]; // 32
  } bcmalloc;                   // 52

typedef void (*logFuncPtr) (char *);

/* public methods: */
void bcmInit(int onoff);
void setLogFunction(logFuncPtr);
// void *malloc(size_t x);
void *BCMMalloc(size_t x, char *file, int line);
//void free(void *x);
void BCMFree(void *x, char *file, int line);
// void *realloc(void *x, size_t y);
void *BCMRealloc(void *x, size_t y, char *file, int line);
// void *calloc(size_t x, size_t y);
void *BCMCalloc(size_t x, size_t y, char *file, int line);
int  checkMemoryBlock(void *m, char *f, int l, char *func);
void randomSet(void *m, void*, int l);

/* internal usage only: */
void *bCheckedMalloc(size_t x, char *file, int line);
void* uncheckedMalloc(size_t x, char *file, int line);
// void bCheckedFree(void *, char *file, int line);
void uncheckedFree(void *, char *file, int line);
void defaultLog(char *msg);
// void *bCheckedRealloc(void *old, size_t x, char *file, int line);
void * uncheckedRealloc(void *old, size_t x, char *file, int line);
void * realmallocFirst(size_t);
#endif
