/*
 * MemCheck, a memory-checker for C/C++.
 * 
 * This file is released under the terms of the GNU General Public License
 * as they appear in http://www.gnu.org/licenses/gpl.txt.
 * 
 * Copyright 2001, 2002 erwin.rieger@ibrieger.de
 */
/****************************************************************/
/* MEMCHECK							*/
/* Memory Checker                                               */
/****************************************************************/

#ifndef _memcheck_h_
#define _memcheck_h_

#ifdef __cplusplus 
	// disable some asm code in bastring.h:
	#undef __i486__
	#undef __i586__
	#undef __i686__
#else
	#ifdef __GNUC__
  	// disable some asm code in selectbits.h:
  	#undef __GNUC__
  	#define __GNUC__ 1
	#endif
#endif

#ifndef __NO_CTYPE
  // disable is<xxx>() macros (isspace() ...)
  #define __NO_CTYPE 1
#endif

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/types.h>
#include <mc_ucl.h>

  #define volatile /* notused */
#ifdef WIN32
  #include <windows.h>
  #include <string.h>
  #include <malloc.h>
#endif

#define MCOUT stderr
#define mcglobalP mcglobal_ptr
#define MCUN 0
#define MCRO 1
#define MCWO 2
#define MCRW 3
// #define MCWO_TO_RW 4
// #define MCWO_TO_RO 5
#define MCNMOPTS "-n -C --defined-only"
// must be multiple of 4:
#define MCHeapBlock (0x1000000)       /* 16 meg */
#define MCBMSIZE  0x400000            /* 16 meg / 4 = 4 meg */
#define MCHeapSegments 256            /* 0x1000000 * 256 = 0x1 0000 0000, 4 gigs */
// #define MCInitialStackSize (60000)
#define MCROBYTE (0x55)
#define MCWOBYTE (0xaa)
#define MCRWBYTE (0xff)
// additional space before main code, space for undef symbols, import stubs
// #define MCMAINSHADOW 5000000

/* default stack right */
// note: must be wo, gcc missing some set_rights ???
#define MCSTACKBYTE MCWOBYTE

/* symbol names */
#ifdef WIN32
  #define SymEtext "__data_start__"
  #define SymEtext2 "_fini"
  #define SymBdata "__data_start__"
  #define SymBBSS "__bss_start__"
  #define SymEBSS "__bss_end__"
#else
  #define SymBtext "_init"
  #define SymBtext2 "_start"
  #define SymEtext "_etext"
  #define SymEtext2 "_fini"
  #define SymBdata "__data_start"
  #define SymBBSS "__bss_start"
  #define SymEBSS "_end"
#endif

// #define MCBM_LEN(x) ((ulong)MCALIGNED(x) ? (x)>>2 : ((x)>>2)+1)
#define MCALIGNED(x) (((ulong)x)==MCALIGN(x))
#define MCALIGN(x) (((ulong)x) & 0xfffffffc)
#define MCALIGN2(x) (((ulong)x) & 0xfffffffe)
#define INTERSECT(a, alen, b, blen) (!(((a+alen)<=b) || ((b+blen)<=a)))

// #define MCINLINE /* notused */
#define MCINLINE inline

#if defined(MEMCHECKDEBUG)
  // debug build
  #include "debugrename.h"
  // #define dr(name) debug_##name
#else
  // #define dr(name) name
#endif

#define BCM_FOR_MC 1
#include <bcm/bcm.h>

#ifdef __cplusplus 
extern "C" {
#endif

typedef unsigned char uchar;
#define _uchar_defined
// typedef unsigned int uint;

/* windows: */
#if defined(__CYGWIN__)
  typedef unsigned long ulong;
  typedef int (*__compar_fn_t) (const void *, const void *);
#endif

typedef struct {
  ulong   addr;
  ulong   end;
  uchar    *bm;
  } mcsegment;

typedef struct {
  // ulong   addr; starts with 0
  // ulong   end; ends with 1 0000 0000
  int     hscount;
  mcsegment segments[MCHeapSegments];
  } mcheapsegment;

typedef struct {
  ulong  addr;
  char  *name;  
  } mcsymtab_list;

typedef struct {
  // ulong  addr;
  bcmalloc  *bcm_alloc;
  } bcmalloc_list;

typedef struct {
  ulong addr;
  int   len;
  char  r;
  } mcright;

typedef struct {
  int           len;
  int           count;
  void          *list;
  int           elsize;
  __compar_fn_t compfunc;
  } mclist;

typedef struct {
  int line;
  int refcount;
  int sum;
  } mcline;

typedef struct mcfile {
  char           *file;
  mcline  *lines;
  int            linescount;
  int            lineslen;
  } mcfile;

typedef struct {
  void *m;  /* muss an erster stelle stehen */
  int  size;
  int  line;
  char *file;
  int  mline;
  char *mfile;
  time_t bd;
  } mcfree_list;

typedef struct {
  ulong handle;  /* muss an erster stelle stehen */
  ulong textstart; // address and length of our memory, needed to reset rights
  // int   len;       // after unloading a shared lib
  ulong datastart;
  ulong bssstart;
  ulong bssend;
  char    *filename;
  int   refcnt;
  int   checked;	// checked code (compiled with -fcheck-memory-usage)
			// 0=unchecked, 1=checked, 2=memcheck itself
  // ures symbols
  // xxx char *unresolved[];
  } mclib_list;

typedef struct mcglob {
  int     abort_ignore;
  int     abort_count;

  int     fileslen;
  int     filescount;
  int     freelistlen;
  int     freelistcount;

  mcheapsegment heap;
  ulong   stacklow;

  mclist  allocs;
  mclist  libs;
  int     dcount;
  int     ncount;
  int     checkstackmode;
  int     checkstackwritemode;
  int     verbose;
  char    *progname;
  char    *checked_libs; // libs compiled with memcheck
  mcfile  *files;
  mcfree_list *freelist;

#ifdef __CYGWIN__
  // ptr to cygwin impure_ptr
  struct _reent *impure_ptr;
#endif
  ulong   main_bp;
  uchar byte_rights[4]; // ={ 0x00, MCROBYTE, MCWOBYTE, MCRWBYTE };
  // ulong textstart;
  // ulong datastart;
  // ulong bssstart;
  // ulong bssend;

  // unchecked libraries:
  // stack of called ucl fcns, array of pointers to func_info's
  // ucl_funcstack_len points to top
  struct func_info *ucl_funcstack[UCLFUNCSTACSIZE];
  unsigned int ucl_funcstack_len;
  mclist  unresolved;
  } mcglob;

// prototypes
void MemCheckInit(void);
void MemCheckClose(int);
void setMCGlobalPtr(mcglob *);
mcglob* getMCGlobalPtr();
void print_location(const void *ptr, size_t len, char right, mcsegment *seg);
int load_libs(const char *name);
void check_free();
void check_allocs();
void printall();
void mcabort(char*);
int get_env_int(char *name, int);
// int get_symtab(const char * name, mclib_list *lib);
int get_symtab(int);
int comp_rights_fast(const void *, const void *);
void check_abort(int bed, char *);
void *list_find(mclist *, void* search);
void *get_sym(void *lib, char *name, const char *filename);
#if defined(MEMCHECKDEBUG)
	ulong getbp1();
#endif
void cpbm(const void *, size_t, void*);
void mergeRights(const void *, size_t, unsigned char);
ulong findTextStart();
void add_lib(mclist*, mclib_list*);
void mcprint_libs(mclist *list);

void chkr_set_right (const void* ptr, size_t len, unsigned char right);
void chkr_check_addr (const void* ptr, size_t len, unsigned char right1);
void chkr_check_addrNoStack (const void* ptr, size_t len, unsigned char right1);
void chkr_copy_bitmap (void* dest, const void* src, size_t orig_len);
void chkr_check_exec (const void *ptr);
void chkr_check_str (const void* ptr, unsigned char right1);

void *MCMalloc(const size_t x, char*, int);
void *MCMalloc_NoInfo(const size_t x);
void MCFree(const void *x, size_t s, char *, int);
void MCFree_NoInfo(const void *x);
void *MCRealloc(void *p, size_t s, char *, int);
void *MCRealloc_NoInfo(void *p, size_t s);
void *MCCalloc(size_t, size_t, char*, int);
void *MCCalloc_NoInfo(size_t, size_t);
char *MCFgets(char *S, int COUNT, FILE *STREAM, char *, int);
char *MCFgets_NoInfo(char *S, int COUNT, FILE *STREAM);
size_t MCStrlen(const char *, char*, int);
size_t MCStrlen_NoInfo(const char *);
char *MCStrcpy(char *TO, const char *FROM, char*, int);
char *MCStrcpy_NoInfo(char *TO, const char *FROM);
char *MCStrcat(char *TO, const char *FROM, char*, int);
char *MCStrcat_NoInfo(char *TO, const char *FROM);
char *MCStrncpy(char *TO, const char *FROM, size_t, char*, int);
char *MCStrncpy_NoInfo(char *TO, const char *FROM, size_t);
int MCStrcmp(const char *S1, const char *S2, char*, int);
int MCStrcmp_NoInfo(const char *S1, const char *S2);
int checknull(const void*, char* func, char* file, int line);

extern void *__real_malloc(size_t);
extern void *__real_memset(void *, int, size_t);
extern void __real_exit(int);
extern int __real_strcmp(const char*, const char *);
extern int __real_fprintf(FILE *, const char*, ...);
extern int __real_sprintf(char*, const char*, ...);
extern void *__real_memcpy(void *, const void*, size_t);
extern void *__real_realloc(void*, size_t);
extern void __real_free(void *);
extern char* __real_strcpy(const char*, const char *);
extern void *__real_dlsym(void *lib, const char *name);
extern char* __real_strcat(char*, const char *);
extern char* __real_strncat(char*, const char *, size_t);
extern void* __real_dlopen(const char*, int flag);
extern int __real_dlclose(void*);
extern time_t __real_time(time_t*);
extern char *__real_fgets(char *s, int count, FILE *f);
extern FILE* __real_fopen(const char*, const char*);
extern size_t __real_fread(void *DATA, size_t SIZE, size_t COUNT, FILE*);
extern char *__real_strncpy(char *TO, const char *FROM, size_t size);
extern void *__real_calloc(size_t, size_t size);
extern int __real_strlen(const char *);
extern size_t __real_strftime(char*, size_t, const char *, const struct tm *);
extern struct tm * __real_localtime(const time_t *);
extern char * __real_dlerror(void);

#ifdef WIN32
  /* include function to get impure_ptr from main module */
  struct _reent * mc_get_impure_ptr();
  #define _impure_ptr mc_get_impure_ptr()
#endif 

// extern void __memcheck_init(void) __attribute__ ((constructor));
// extern void __memcheck_init(void);
// extern void __memcheck_close(void) __attribute__ ((destructor));
 
// void mcset_progname(char *progname); // xxx fuer mcstartup++
char* get_env_string(char *name, char *); // dito
char * mcget_progname(int); // xxx fuer mcstartup++

// MCFUNC void *get_sym(void *lib, char *name, const char *filename);
// MCFUNC void add_symtab(mcsymtab_list *st);
// MCFUNC int comp_symtab(const void *, const void *);
// MCFUNC int comp_symtab_name(const void *, const void *);
// MCFUNC int len_from_symtab(const mcsymtab_list *s);
// MCFUNC mcsymtab_list * find_sym_name(const char *);
// MCFUNC void check_one_free(mcfree_list *f, int);
// MCFUNC void printmem(unsigned char *mm, int size);
// MCFUNC void add_alloc(mcheapsegment *, const bcmalloc *);
// MCFUNC void init_mcglob(mcglob *);
// MCFUNC void add_lib(mclist*, mclib_list*);
// MCFUNC int del_lib(mclist*, void*);
// MCFUNC void list_append(mclist *, void* newelement);
// void mc_print(void);
// void mcobserve_addr(void **addr, void *value);
// void mcregnew(int, char*);
// void mcregdelete(int, char*);

#if defined(MEMCHECK) && !defined(MEMCHECKDEBUG)
  // rename some functions (malloc, free ...) to get line and file information:
  #include "memcheck_macros.h"
#endif

#ifdef __cplusplus 
  }
#endif

#endif
