/*
 * MemCheck, a memory-checker for C/C++.
 * 
 * This file is released under the terms of the GNU General Public License
 * as they appear in http://www.gnu.org/licenses/gpl.txt.
 * 
 * Copyright 2001, 2002 erwin.rieger@ibrieger.de
 */
/****************************************************************/
/* MEMCHECK							*/
/* Modul zur Speicherverwaltungs - Ueberwachung		        */
/****************************************************************/

#ifndef _memcheck_macros_h_
#define _memcheck_macros_h_

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#ifdef WIN32
  #include <winnt.h>
  #include <windows.h>
  #include <windef.h>
  #include <wincon.h>
  #include <basetyps.h>
  #include <winbase.h>
  #include <wingdi.h>
  #include <winuser.h>
#endif

#ifndef __cplusplus

// asm can't be used with -fcheck-memory-usage:
#define __asm__(x) fprintf(stderr, "Memcheck: warning, asm-code x not executed!\nIn %s, line %d", __FILE__, __LINE__);
#define asm(x) fprintf(stderr, "Memcheck: warning, asm-code x not executed!\nIn %s, line %d", __FILE__, __LINE__);
#define __asm(x) fprintf(stderr, "Memcheck: warning, asm-code x not executed!\nIn %s, line %d", __FILE__, __LINE__);
// linux:
#define __NO_MATH_INLINES 1

/* cfree is an old name for free: */
#define cfree free


#if !defined(MEMCHECKDEBUG)

#ifdef malloc
  #undef malloc
#endif
#define MCMalloc_NoInfo(x) MCMalloc(x, __FILE__, __LINE__)

#ifdef free
  #undef free
#endif
#define MCFree_NoInfo(x)   MCFree(x, 0, __FILE__, __LINE__)

#ifdef realloc
  #undef realloc
#endif
#define MCRealloc_NoInfo(x, y)   MCRealloc(x, y, __FILE__, __LINE__)

#ifdef calloc
  #undef calloc
#endif
#define MCCalloc_NoInfo(x, y)   MCCalloc(x, y, __FILE__, __LINE__)

#ifdef fgets
  #undef fgets
#endif
#define MCFgets_NoInfo(x, y, z)   MCFgets(x, y, z, __FILE__, __LINE__)

#ifdef strlen
  #undef strlen
#endif
#define MCStrlen_NoInfo(x) MCStrlen(x, __FILE__, __LINE__)

#ifdef strcmp
  #undef strcmp
#endif
#define MCStrcmp_NoInfo(x, y) MCStrcmp(x, y, __FILE__, __LINE__)

#ifdef strcpy
  #undef strcpy
#endif
#define MCStrcpy_NoInfo(x, y) MCStrcpy(x, y, __FILE__, __LINE__)

#ifdef strcat
  #undef strcat
#endif
#define MCStrcat_NoInfo(x, y) MCStrcat(x, y, __FILE__, __LINE__)

#ifdef strncpy
  #undef strncpy
#endif
#define MCStrncpy_NoInfo(x, y, z) MCStrncpy(x, y, z, __FILE__, __LINE__)

#define malloc MCMalloc_NoInfo
#define free   MCFree_NoInfo
#define realloc   MCRealloc_NoInfo
#define calloc   MCCalloc_NoInfo
#define fgets   MCFgets_NoInfo
#define strlen MCStrlen_NoInfo
#define strcmp MCStrcmp_NoInfo
#define strcpy MCStrcpy_NoInfo
#define strcat MCStrcat_NoInfo
#define strncpy MCStrncpy_NoInfo

#endif  // #if !defined(MEMCHECKDEBUG)

#endif /* __cplusplus */

#endif
