/*
 * MemCheck, a memory-checker for C/C++.
 * 
 * This file is released under the terms of the GNU General Public License
 * as they appear in http://www.gnu.org/licenses/gpl.txt.
 * 
 * Copyright 2001, 2002 erwin.rieger@ibrieger.de
 */
/* stolen from checker 0.9.9.1 ... */

/* Check arguments for printf, scanf ...
   Copyright 1995, 1996 Tristan Gingold
		  Written December 1995 by Tristan Gingold

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of the
License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License 
along with this program; see the file COPYING.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.

The author may be reached by US/French mail:
		Tristan Gingold 
		8 rue Parmentier
		F-91120 PALAISEAU
		FRANCE
*/


#include <memcheck.h>

#include <stdarg.h>
#include <ctype.h>
#include <sys/types.h>
// #define NEED_CHKR_VA_ARG_ADDR
// #include "machine.h"
#include "check-printf.h"
// #include "checker_api.h"
#include <assert.h>

#ifndef chkr_va_arg_addr
#define chkr_va_arg_addr &va_arg
#endif

#define	LONGINT  1
#define	SHORTINT 2
#define LONGDBL  4
#define WIDTH    8
#define SUPPRESS 16

#define CONV_CHAR 1
#define CONV_INT  2
#define CONV_STRING 3
#define CONV_DOUBLE 4
#define CONV_POINTER 5

/* Number of chars needed to write a decimal number with type t.  */
// #define NBR_CHARS_FOR_TYPE(t) (sizeof(t) * 3)
#define MAXPRINTFSTR 102400	// hoping, no one will print more than 100 k to a string ...
#define MAXFORMATSTR 10240 	// and not more than 10 k format
static int charsNeeded(char *typ, long ival, long double dval, const char *startfmt, \
  const char *fmt) {
  char tmp[MAXPRINTFSTR];
  char realfmt[MAXFORMATSTR];
  const char *tmpp;
  int len=0;

  // for (tmpp=startfmt; tmpp<fmt; tmpp++) realfmt[tmpp-startfmt]=*tmpp;
  for (tmpp=startfmt; tmpp<fmt; tmpp++) {
	realfmt[tmpp-startfmt]=*tmpp;
	assert(++len<MAXFORMATSTR);
	}
  realfmt[tmpp-startfmt]='\0';
  // xprintf(MCOUT, "chars needed: %s, %s, %ld, %Lf\n", typ, realfmt, ival, dval);

  if (!__real_strcmp(typ, "%ld")) {
    if ((len=__real_sprintf(tmp, realfmt, ival))<=0) {
      __real_fprintf(MCOUT, "cannot convert: %ld, '%s'\n", ival, realfmt);
      mcabort("in charsNeeded");
      }
    }
  else if (!__real_strcmp(typ, "%h")) {
    if ((len=__real_sprintf(tmp, realfmt, (short)ival))<=0) {
      __real_fprintf(MCOUT, "cannot convert: %hd, '%s'\n", (short)ival, realfmt);
      mcabort("in charsNeeded");
      }
    }
  else if (!__real_strcmp(typ, "%g")) {
    if ((len=__real_sprintf(tmp, realfmt, (double)dval))<=0) {
      __real_fprintf(MCOUT, "cannot convert: %g, '%s'\n", (double)dval, realfmt);
      mcabort("in charsNeeded");
      }
    }
  else if (!__real_strcmp(typ, "%Lg")) {
    if ((len=__real_sprintf(tmp, realfmt, dval))<=0) {
      __real_fprintf(MCOUT, "cannot convert: %Lg, '%s'\n", dval, realfmt);
      mcabort("in charsNeeded");
      }
    }
  else {
    // %d
    if ((len=__real_sprintf(tmp, realfmt, (int)ival))<=0) {
      __real_fprintf(MCOUT, "cannot convert: %d, '%s'\n", (int)ival, realfmt);
      mcabort("in charsNeeded");
      }
    // xprintf(MCOUT, "chars needed dez vor int %d: '%s', %d\n", (int)ival, realfmt, len);
    }
  // xprintf(MCOUT, "chars needed: '%s', %d\n", realfmt, len);
  assert(len<MAXPRINTFSTR);
  return len;
  }

static size_t mystrnlen (const char *s, size_t n) {
  size_t res = 0;
  while (*s++ && res < n)
    res++;
  if (res < n)
    res++;
  return res;
}

#ifdef CHECK_INCOMING_ARGS
  #undef CHECK_INCOMING_ARGS
  #define CHECK_INCOMING_ARGS 1
#else
  #define CHECK_INCOMING_ARGS 0
#endif

#define CHKR_WO MCWO
#define CHKR_RO MCRO
#define CHKR_RW MCRW
#define CHKR_TW MCWO
#define CHKR_MW MCWO
#define stubs_chkr_set_right chkr_set_right

#define stubs_chkr_check_addr(a, b, c, d) chkr_check_addr(a, b, c)
#define stubs_chkr_check_str(a, b, c) MCStrlen(a, c, 0)

/* TYPE is either TYPE_PRINTF, TYPE_PRESCANF or a positive number of args
   that were written.  */
/* Return the maximum of bytes written.  */
static int
check_printf_format (char const *name, char const *fmt, va_list ap, int type, int va)
{
  int flags;			/* flags as above */
  char c;
  const char *startfmt=fmt;
  int width, w;
  int conv;
  int num;
  int len;

  num = 0;
  len = 1; /* because of '\0'.  */
  
  /* Scan the format for conversions (`%' character). */
  while (1)
    {
      while (*fmt && *fmt != '%')
	{
	  len++;
	  fmt++;
	}

      if (*fmt == '\0')
	return len;

      startfmt=fmt;
      fmt++;			/* skip over '%' */

      flags = 0;
      width = 0;
      conv = 0;

      if (type >= 0 && type == num)
	{
	  /* All the args where checked.  */
	  return 0;
	}
      else
        num++;

    rflag:
      switch ((c = *fmt++))
	{
	case '%':
	  len++;
          startfmt=fmt;
	  continue;	/* ignore this char */
	case ' ':
	case '#':
	case '-':
	case '+':
	  goto rflag;
	case '*':
	  if (type == TYPE_PRINTF)
	    {
	      int *addr = chkr_va_arg_addr (ap, int);
	      if (va || CHECK_INCOMING_ARGS)
	        stubs_chkr_check_addr(addr, sizeof (int), CHKR_RO, "%*");
	      flags |= WIDTH;
	      width += *addr;
	    }
	  else
	    flags |= SUPPRESS;
	  goto rflag;
	case '.':
	  if (type == TYPE_PRINTF && *(fmt+1) == '*')
	    {
	      int *addr = chkr_va_arg_addr (ap, int);
	      if (va || CHECK_INCOMING_ARGS)
	        stubs_chkr_check_addr(addr, sizeof (int), CHKR_RO, "%*");
	      width += *addr;
	      flags |= WIDTH;
				fmt++;
	    }
	  goto rflag;
	case '0':
	  goto rflag;
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
			w=(c-'0');
			c=*fmt++;
	    while (isdigit ((unsigned char)c)) {
				w = w * 10 + (c - '0');
				c = *fmt++;
	      }
	    width+=w;
	    fmt--;
	    flags |= WIDTH;
	  goto rflag;
 	case 'L':
	  flags |= LONGDBL;
	  goto rflag;
	case 'h':
	  flags |= SHORTINT;
	  goto rflag;
	case 'l':
	  flags |= LONGINT;
	  goto rflag;
	case 'c':
	  conv = CONV_CHAR;
	  break;
	case 'D':
	  flags |= LONGINT;
	case 'd':
	case 'i':
	  conv = CONV_INT;
	  break;
	case 'e':
	case 'E':
	case 'f':
	case 'F':
	case 'g':
	case 'G':
	  conv = CONV_DOUBLE;
	  break;
	case 'n':
	  if (type != TYPE_PRESCANF)
	    {
	      if (flags & LONGINT)
		{
		  long **addr;
		  addr = chkr_va_arg_addr (ap, long *);
		  if (va || CHECK_INCOMING_ARGS)
		    stubs_chkr_check_addr(addr, sizeof (long *), CHKR_RO, "%n");
		  stubs_chkr_check_addr(*addr, sizeof (long), CHKR_WO, "%n");
		}
	      else if (flags & SHORTINT)
		{
		  short **addr;
		  addr = chkr_va_arg_addr (ap, short *);
		  if (va || CHECK_INCOMING_ARGS)
		    stubs_chkr_check_addr(addr, sizeof (short *), CHKR_RO, "%n");
		  stubs_chkr_check_addr(*addr, sizeof (short), CHKR_WO, "%n");
		}
	      else
		{
		  int **addr;
		  addr = chkr_va_arg_addr (ap, int *);
		  if (va || CHECK_INCOMING_ARGS)
		    stubs_chkr_check_addr(addr, sizeof (int *), CHKR_RO, "%n");
		  stubs_chkr_check_addr(*addr, sizeof (int), CHKR_WO, "%n");
		}
	    }
	  continue;		/* no output */
	case 'O':
	  flags |= LONGINT;
	case 'o':
	  conv = CONV_INT;
	  break;
	case 'p':
	  conv = CONV_POINTER;
	  break;
	case 's':
	  conv = CONV_STRING;
	  break;
	case 'U':
	  flags |= LONGINT;
	case 'u':
	  conv = CONV_INT;
	  break;
	case 'X':
	case 'x':
	  conv = CONV_INT;
	  break;
	case '[':
	  if (type != TYPE_PRINTF)
	    {
	      if (*fmt++ == '^')
		fmt++;		/* skip the character after the '^' */
	      while (*fmt && *fmt != ']')
		fmt++;
	      if (!*fmt)
	        {
	          // chkr_report (M_C_FUN_LB_ET);
	          // chkr_printf ("%s: Bad format after '['\n", name);
	          // chkr_disp_call_chain ();
		  __real_fprintf(MCOUT, "%s: Bad format after '['\n", name);
		  mcabort("error in check_printf, see messages above");
		}
	      fmt++;
	      goto rflag;
	      // break;
	    }
	  /* FALL THROUGH */
	default:
	  // chkr_report (M_C_FUN_LB_ET);
	  // chkr_printf ("%s: Unknown type '%c'\n", name, c);
	  // chkr_disp_call_chain ();
	  __real_fprintf(MCOUT, "%s: Unknown type '%c'\n", name, c);
	  mcabort("error in check_printf, see messages above");
	}

      if (type == TYPE_PRINTF)
	switch (conv)
	  {
	  case CONV_CHAR:
	    {
	      int *addr;
	      addr = chkr_va_arg_addr (ap, int);
	      if (va || CHECK_INCOMING_ARGS)
	        stubs_chkr_check_addr(addr, sizeof (int), CHKR_RO, "%c");
	      len++;
	    }
	    break;
	  case CONV_INT:
	    if (flags & LONGINT) {
				long *addr;
				addr = chkr_va_arg_addr (ap, long);
				if (va || CHECK_INCOMING_ARGS)
					stubs_chkr_check_addr(addr, sizeof (long), CHKR_RO, "%ld");
				// len += NBR_CHARS_FOR_TYPE(long);
				len += (width>0) ? width : charsNeeded("%ld", *addr, 0, startfmt, fmt);
	      }
	    else if (flags & SHORTINT) {
				int *addr;
				addr = chkr_va_arg_addr (ap, int);
				if (va || CHECK_INCOMING_ARGS)
					stubs_chkr_check_addr(addr, sizeof (short), CHKR_RO, "%h");
				// len += NBR_CHARS_FOR_TYPE(short int);
				len += (width>0) ? width : charsNeeded("%h", (long)*addr, 0, startfmt, fmt);
	      }
	    else {
				int *addr;
				addr = chkr_va_arg_addr (ap, int);
					if (va || CHECK_INCOMING_ARGS)
				stubs_chkr_check_addr(addr, sizeof (int), CHKR_RO, "%d");
				// len += NBR_CHARS_FOR_TYPE(int);
				len += (width>0) ? width : charsNeeded("%d", (long)*addr, 0, startfmt, fmt);
	      }
	    break;
	  case CONV_DOUBLE:
	    {
              if (flags & LONGDBL) {
	        long double *addr;
	        addr = chkr_va_arg_addr (ap, long double);
	        if (va || CHECK_INCOMING_ARGS)
	          stubs_chkr_check_addr(addr, sizeof (long double), CHKR_RO, "%g");
	        len += (width>0) ? width : charsNeeded("%Lg", 0, *addr, startfmt, fmt);
                }
              else {
	        double *addr;
	        addr = chkr_va_arg_addr (ap, double);
	        if (va || CHECK_INCOMING_ARGS)
	          stubs_chkr_check_addr(addr, sizeof (double), CHKR_RO, "%g");
	        // len += NBR_CHARS_FOR_TYPE(double) + 2;
	        len += (width>0) ? width : charsNeeded("%g", 0, (long double)*addr, startfmt, fmt);
                }
	    }
	    break;
	  case CONV_POINTER:
	    {
	      void **addr;
	      addr = chkr_va_arg_addr (ap, void *);
	      if (va || CHECK_INCOMING_ARGS)
	        stubs_chkr_check_addr(addr, sizeof (void *), CHKR_RO, "%p");
	      len += sizeof (char *) * 2 + 2; /* 0x */
	    }
	    break;
	  case CONV_STRING: {
      char **addr;
      addr = chkr_va_arg_addr (ap, char *);
      if (va || CHECK_INCOMING_ARGS)
        stubs_chkr_check_addr(addr, sizeof (char *), CHKR_RO, "%s");
      if (*addr) {
		  	if (flags & WIDTH) {
		      if (width) {
			  		width = mystrnlen (*addr, width);
			  		stubs_chkr_check_addr(*addr, width, CHKR_RO, "%s");
					}
		      len += width;
		    }
		  	else {
	      	stubs_chkr_check_str(*addr, CHKR_RO, "%s");
	      	len += strlen (*addr);
		    	}
				}
      else {
		  	/* printf ("%s", NULL) usually prints (null).  */
		  	len += 6;
	      // chkr_report (M_C_FUN_LB_ET);
	      // chkr_printf ("%s: NULL argument for format `%%s'\n", name);
	      // chkr_disp_call_chain ();
	      __real_fprintf(MCOUT, "%s: NULL argument for format `%%s'\n", name);
		  	mcabort("error in check_printf, see messages above");
			}
	    }
	    break;
	  }
    else if (type == TYPE_PRESCANF && !(flags & SUPPRESS))
	switch (conv)
	  {
	  case CONV_CHAR:
	    {
	      char **addr;
	      addr = chkr_va_arg_addr (ap, char*);
	      if (va || CHECK_INCOMING_ARGS)
	        stubs_chkr_check_addr(addr, sizeof (char *), CHKR_RO, "%c");
	      stubs_chkr_check_addr(*addr, width ? width * sizeof(char) : sizeof(char), CHKR_MW, "%c");
	    }
	    break;
	  case CONV_INT:
	    if (flags & LONGINT)
	      {
		long **addr;
		addr = chkr_va_arg_addr (ap, long*);
		if (va || CHECK_INCOMING_ARGS)
		  stubs_chkr_check_addr(addr, sizeof (long*), CHKR_RO, "%ld");
		stubs_chkr_check_addr(*addr, sizeof (long), CHKR_MW, "%ld");
	      }
	    else if (flags & SHORTINT)
	      {
		int **addr;
		addr = chkr_va_arg_addr (ap, int*);
		if (va || CHECK_INCOMING_ARGS)
		  stubs_chkr_check_addr(addr, sizeof (short*), CHKR_RO, "%h");
		stubs_chkr_check_addr(*addr, sizeof (short), CHKR_MW, "%h");
	      }
	    else
	      {
		int **addr;
		addr = chkr_va_arg_addr (ap, int*);
		if (va || CHECK_INCOMING_ARGS)
		  stubs_chkr_check_addr(addr, sizeof (int*), CHKR_RO, "%d");
		stubs_chkr_check_addr(*addr, sizeof (int), CHKR_MW, "%d");
	      }
	    break;
	  case CONV_DOUBLE:
	    {
              if (flags & LONGDBL) {
	        long double **addr;
	        addr = chkr_va_arg_addr (ap, long double*);
	        if (va || CHECK_INCOMING_ARGS)
	          stubs_chkr_check_addr(addr, sizeof (long double*), CHKR_RO, "%e");
	        stubs_chkr_check_addr(*addr, sizeof (long double), CHKR_MW, "%e");
                } 
              else {
	        double **addr;
	        addr = chkr_va_arg_addr (ap, double*);
	        if (va || CHECK_INCOMING_ARGS)
	          stubs_chkr_check_addr(addr, sizeof (double*), CHKR_RO, "%e");
	        stubs_chkr_check_addr(*addr, sizeof (double), CHKR_MW, "%e");
                }
	    }
	    break;
	  case CONV_POINTER:
	    {
	      void ***addr;
	      addr = chkr_va_arg_addr (ap, void **);
	      if (va || CHECK_INCOMING_ARGS)
	        stubs_chkr_check_addr(addr, sizeof (void **), CHKR_RO, "%p");
	      stubs_chkr_check_addr(*addr, sizeof (void *), CHKR_MW, "%p");
	    }
	    break;
	  case CONV_STRING:
	    {
	      char **addr;
	      addr = chkr_va_arg_addr (ap, char *);
	      if (va || CHECK_INCOMING_ARGS)
	        stubs_chkr_check_addr(addr, sizeof (char *), CHKR_RO, "%s");
	      if (!width)
	        {
	          // chkr_report (M_C_FUN_LB_ET);
	          // chkr_printf("%s: No width field for string.\n", name);
	          // chkr_disp_call_chain ();
	          __real_fprintf(MCOUT, "%s: No width field for string.\n", name);
		  mcabort("error in check_printf, see messages above");
	        }
	      else
	        stubs_chkr_check_addr(*addr, width + 1, CHKR_MW, "%s");
	    }
	    break;
	  }
      else if (!(flags & SUPPRESS)) {
	switch (conv)
	  {
	  case CONV_CHAR:
	    {
	      char **addr;
	      addr = chkr_va_arg_addr (ap, char*);
	      stubs_chkr_set_right (*addr, width ? width * sizeof(char) : sizeof(char), CHKR_RW);
	    }
	    break;
	  case CONV_INT:
	    if (flags & LONGINT)
	      {
		long **addr;
		addr = chkr_va_arg_addr (ap, long*);
		stubs_chkr_set_right (*addr, sizeof (long), CHKR_RW);
	      }
	    else if (flags & SHORTINT)
	      {
		int **addr;
		addr = chkr_va_arg_addr (ap, int*);
		stubs_chkr_set_right (*addr, sizeof (short), CHKR_RW);
	      }
	    else
	      {
		int **addr;
		addr = chkr_va_arg_addr (ap, int*);
		stubs_chkr_set_right (*addr, sizeof (int), CHKR_RW);
	      }
	    break;
	  case CONV_DOUBLE:
	    {
              if (flags & LONGDBL) {
	        long double **addr;
	        addr = chkr_va_arg_addr (ap, long double*);
	        stubs_chkr_set_right (*addr, sizeof (long double), CHKR_RW); 
                }
              else {
	        double **addr;
	        addr = chkr_va_arg_addr (ap, double*);
	        stubs_chkr_set_right (*addr, sizeof (double), CHKR_RW); 
                }
	    }
	    break;
	  case CONV_POINTER:
	    {
	      void ***addr;
	      addr = chkr_va_arg_addr (ap, void **);
	      stubs_chkr_set_right (*addr, sizeof (void *), CHKR_RW);
	    }
	    break;
	  case CONV_STRING:
	    {
	      char **addr;
	      int n;
	      addr = chkr_va_arg_addr (ap, char *);
		  	if (flags & WIDTH) {
					assert(width);
					#ifdef MEMCHECKDEBUG
    					mergeRights(*addr, width, MCRO); // memory not initialized by vfscanf
					#endif
		      // if (width) {
			  		n=mystrnlen(*addr, width)+1;
	          // xprintf(MCOUT, "width: %d mylen: %d\n", width, n);
						// mcabort("error in check_printf, see messages above");
					// }
        }
				else {
					#ifdef MEMCHECKDEBUG
							assert(0);
    					//mergeRights(*addr, sizeof(long)*2, MCRW);
					#endif
	      	n = strlen (*addr) + 1;	/* FIXME */
				}
	      stubs_chkr_set_right (*addr, n * sizeof (char), CHKR_RW);
	    }
	    break;
	  }
	}
	else {
		mcabort("internal error in check_printf_format");
	}
    }
}

int __wrap_sprintf (char *str, const char *format, ... ) {
  va_list param;
  int res;
  int len;
  
  va_start (param, format);
  stubs_chkr_check_str(format, CHKR_RO, "format");
  len = check_printf_format ("sprintf", format, param, TYPE_PRINTF, 1);
  // xprintf(MCOUT, "checker:sprintf need %d chars (%s)\n", len, format);
  stubs_chkr_check_addr(str, len, CHKR_TW, "str");
  res = vsprintf (str, format, param);
  if (res == (int)str)
    stubs_chkr_set_right (str, strlen (str) + 1, CHKR_RW);
  else if (res >= 0) {
    stubs_chkr_set_right (str, res + 1, CHKR_RW);
    // xprintf(MCOUT, "res sprintf: %s, %s\n", format, str);
    }
  return res;
}

#if defined(__CYGWIN__) || defined(MEMCHECKDEBUG)
  // cygwin, has no vsscanf
  // extern int vsscanf(const char*, const char*, ...);
  // int vsscanf(const char *str, const char *format, va_list params);
  #include <mcvsscanf.h>
#endif
  
int __wrap_sscanf (const char *str, const char *format, ... ) {
  va_list param;
  int n;
  
  va_start (param, format);
  stubs_chkr_check_str(str, CHKR_RO, "str");
  stubs_chkr_check_str(format, CHKR_RO, "format");
  check_printf_format ("sscanf", format, param, TYPE_PRESCANF, 1);
  n = vsscanf (str, format, param);
  if (n>=0)
    check_printf_format ("sscanf", format, param, n, 0);
  va_end (param);
  return n;
}

// extern int __real_fscanf(FILE *, const char *, ...);
int __wrap_fscanf (FILE *fp, const char *format, ... ) {
  va_list param;
  int n;
  
  va_start (param, format);
  stubs_chkr_check_str(format, CHKR_RO, "format");
  check_printf_format ("fscanf", format, param, TYPE_PRESCANF, 1);
  n = vfscanf (fp, format, param);
  if (n>=0)
    check_printf_format ("fscanf", format, param, n, 0);
  va_end (param);
  return n;
}

