/*
 * MemCheck, a memory-checker for C/C++.
 * 
 * This file is released under the terms of the GNU General Public License
 * as they appear in http://www.gnu.org/licenses/gpl.txt.
 * 
 * Copyright 2001, 2002 erwin.rieger@ibrieger.de
 */
//
// 
//

// Here's some code that will walk the objects that make up a dynamic (32-bit)
// process and print common symbol information.  NOTE - the only symbol table
// avialable withina process is the .dynsym, which is mapped as part of the text
// segment.  The.symtab is larger, containing locals (and any symbols that have
// been scoped), and is only available from the file - that's why debuggers
// typically need the original files to provide maximum symbol information.
// A more complete exmaple might be a good condidate for /usr/demo/ELF.
// Have fun.


#include <dlfcn.h>
#include <link.h>
#include <libelf/libelf.h>
#include <stdio.h>
#include <assert.h>
#include <memcheck.h>

static void
display(Elf32_Sym * sym, char * str, unsigned long * hsh, unsigned long base)
{
	unsigned long	ndx, cnt = hsh[1];

	/*
	 * Traverse the objects symbol table (note, the number of symbols is
	 * deduced from the hash table).
	 */
	for (sym++, ndx = 1; ndx < cnt; sym++, ndx++) {
		char *		type, * bind;
		unsigned long	bgn;

		if ((sym->st_name == SHN_UNDEF) || (sym->st_size==0)) continue;

		/*
		 * Determine symbols binding.
		 */
		assert(ELF32_ST_BIND(sym->st_info)!=STB_LOCAL);
		if (ELF32_ST_BIND(sym->st_info) == STB_WEAK)
			bind = "WEAK";
		else
			bind = "GLOB";

		/*
		 * Determine symbols type.
		 */
		if (ELF32_ST_TYPE(sym->st_info) == STT_FUNC)
			type = "FUNC";
		else if (ELF32_ST_TYPE(sym->st_info) == STT_OBJECT)
			type = "OBJT";
		else assert(0);

		/*
		 * Determine symbols begin memory address.
		 */
		bgn = sym->st_value + base;

		printf("\t%8.8lx - %8.8lx\t%s %s\t%s\n", bgn, sym->st_size, type, bind, str + sym->st_name);
	}
}

int dl_get_objects() {
	struct link_map *	lmp;
	struct r_debug	*rdebug=NULL;
	Elf32_Dyn *	dyn;
	mclib_list	lib;

	/*
	 * Obtain the applications link-map.
	 */
	for (dyn = _DYNAMIC; dyn->d_tag != DT_NULL; ++dyn)
		if (dyn->d_tag == DT_DEBUG)
			rdebug = (struct r_debug *) dyn->d_un.d_ptr;
	lmp=rdebug->r_map;

	/*
	 * Traverse each library of the process.
	 */
	while (lmp) {
		if (strlen(lmp->l_name)==0) {
			lib.filename=mcget_progname(1);
			// get base addr from address of dynamic tab
			// lib.textstart=findTextStart() & 0xff000000;
			lib.textstart=(ulong)lmp->l_ld & 0xfff00000;
			}
		else {
			lib.filename=lmp->l_name;
			lib.textstart=lmp->l_addr;
			}

		printf("base of lib %s at: %x\n", lib.filename, lmp->l_addr);
		lib.handle=(ulong)lmp;
		lib.datastart=0;
		lib.bssstart=0;
		lib.bssend=0;
		lib.refcnt=1;
		lib.checked=0;
		add_lib(&(getMCGlobalPtr()->libs), &lib);
		lmp = lmp->l_next;
	}
	return 1;
}

int dl_get_symtab() {
	void *		handle;
	struct link_map *	lmp;
	struct r_debug	*rdebug=NULL;
	Elf32_Dyn *	dyn;
	unsigned int	bas = 0;

	/*
	 * Obtain the applications link-map.
	 */
	for (dyn = _DYNAMIC; dyn->d_tag != DT_NULL; ++dyn)
		if (dyn->d_tag == DT_DEBUG)
			rdebug = (struct r_debug *) dyn->d_un.d_ptr;
	lmp=rdebug->r_map;

	/*
	 * Traverse each library of the process.
	 */
	while (lmp) {
		Elf32_Ehdr *	ehdr;
		Elf32_Phdr *	phdr;
		int		cnt;
		char *		sym = 0, * str = 0, * hsh = 0;

		if (strlen(lmp->l_name)==0)
			(void) printf("\nInspecting MAIN\n");
		else
			(void) printf("\nInspecting %x, %s\n", lmp->l_name, lmp->l_name);
		printf("base of lib at: %x\n", lmp->l_addr);
		printf("dyn section at: %x\n", lmp->l_ld);
		dyn=lmp->l_ld;

		/*
		 * Traverse dynamic section looking for SYMTAB,
		 * STRTAB and HASH.
		 */
		for (; dyn->d_tag != DT_NULL; dyn++) {
			if (dyn->d_tag == DT_SYMTAB)
				sym = (char *)dyn->d_un.d_ptr;
			else if (dyn->d_tag == DT_STRTAB)
				str = (char *)dyn->d_un.d_ptr;
			else if (dyn->d_tag == DT_HASH)
				hsh = (char *)dyn->d_un.d_ptr;
		}

		printf("sym: %x, str: %x, hsh: %x\n", sym, str, hsh);
		if (sym && str && hsh) {
			/*
			 * Establish each sections memory
			 * address.
			 */
			if (lmp->l_addr != 0) {
				hsh += lmp->l_addr;
			}
			display((Elf32_Sym *)sym, str, (unsigned long *)hsh, lmp->l_addr);
		}
		lmp = lmp->l_next;
	}

	return (0);
}

