/*
 * MemCheck, a memory-checker for C/C++.
 * 
 * This file is released under the terms of the GNU General Public License
 * as they appear in http://www.gnu.org/licenses/gpl.txt.
 * 
 * Copyright 2001, 2002 erwin.rieger@ibrieger.de
 */

//#if defined(_memcheck_h_)
  //#error "memcheck.c: memcheck.h already included !"
//#endif

#include "memcheck.h"

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <malloc.h>
#include <limits.h>
#include <time.h>
#include <ctype.h>
#include <signal.h>
#include <dlfcn.h>
#include <dirent.h>
#include <sys/stat.h>
#include <time.h>
#include <assert.h>
#include <sys/ptrace.h>
#include <sys/wait.h>

#if defined(WIN32)
	#include <windows.h>
	#include <winnt.h>
#else
	extern char *__progname;
	extern char *__progname_full;
#endif

#define MAXBUFFER 1024
#if defined(AIXV3)
extern char * sys_errlist[];
#endif

#define CheckRights 1

#ifdef WIN32
static struct _reent *my_impure_ptr;
#endif

// singleton:
static mcglob mcglobal;

static void init_mcglob(mcglob *g) {
  int aufdemstack;

  g->stacklow=(ulong)&aufdemstack;
  for (aufdemstack=0; aufdemstack<MCHeapSegments; aufdemstack++) {
    g->heap.segments[aufdemstack].bm=NULL;
    g->heap.segments[aufdemstack].addr=aufdemstack * MCHeapBlock;
    g->heap.segments[aufdemstack].end=(aufdemstack * MCHeapBlock) + MCHeapBlock;
    }

  g->allocs.len=0;
  g->allocs.count=0;
  g->allocs.list=NULL;
  g->allocs.elsize=sizeof(bcmalloc_list);
  g->allocs.compfunc=mc_comp_int;

  g->libs.len=0;
  g->libs.count=0;
  g->libs.list=NULL;
  g->libs.elsize=sizeof(mclib_list);
  g->libs.compfunc=mc_comp_int;

  g->unresolved.len=0;
  g->unresolved.count=0;
  g->unresolved.list=NULL;
  g->unresolved.elsize=sizeof(char*);
  g->unresolved.compfunc=mc_comp_int;

  g->fileslen=0;
  g->filescount=0;

  g->freelistlen=0;
  g->freelistcount=0;

  g->ncount=0;
  g->dcount=0;
  g->verbose=2;
  // g->progname=NULL;

  g->abort_ignore=0;
  g->abort_count=0;
  g->files=NULL;
  g->freelist=NULL;

  g->byte_rights[0]=0x00;
  g->byte_rights[1]=MCROBYTE;
  g->byte_rights[2]=MCWOBYTE;
  g->byte_rights[3]=MCRWBYTE;
  // g->textstart=0;  // is set in memcheck.c:get_symtab
  g->checkstackwritemode=0;
  g->checkstackmode=0;

  g->ucl_funcstack_len=0;
  }

// void mcset_progname(char *progname) {
  // __real_fprintf(stderr, "progname is: %s\n", progname);
  // mcglobal.progname=progname;
  // }
char* mcget_progname(int full) {
  if (full)
	return __progname_full;
  return __progname;
  }

void MemCheckInit() {
  static int    initialized=0;
  int i, mode;
#ifdef __CYGWIN__
  char buffer[1024];
#endif

  __real_fprintf(MCOUT, "MemCheckInit called ...\n");
  if (initialized++) {
    __real_fprintf(MCOUT, "\nMemCheckInit(), warning: multiple init\n");
    return;
    }

  init_mcglob(&mcglobal);

  // give access to global data to lib part:
  setMCGlobalPtr(&mcglobal);

  mcglobal.abort_ignore=get_env_int("MCIGNORE", 0);
  mcglobal.verbose=get_env_int("MCVERBOSE", 2);
  #if ! defined(MEMCHECKDEBUG)
    mode=get_env_int("MCMODE", 0xfffffff);
  #else
    mode=get_env_int("DEBMCMODE", 0xfffffff);
  #endif
  mcglobal.checked_libs=get_env_string("MCCHECKEDLIBS", NULL);
  // bit 0: check stack write
  // bit 1: check stack
  if (mode & 1) mcglobal.checkstackwritemode=1;
  if (mode & 2) mcglobal.checkstackmode=1;

  if (!dl_get_objects()) {
    __real_fprintf(MCOUT, "Cannot get shared object deps !\n");
    __real_exit(-1);
    }
  mcprint_libs(&mcglobal.libs);
  for (i=0; i<mcglobal.libs.count; i++) {
	if (!get_symtab(i)) { 
		__real_fprintf(MCOUT, "Cannot get Symbols from %s (nm) !\n", mcglobal.progname);
		__real_exit(-1);
	}
  }
  mcprint_libs(&mcglobal.libs);

  assert(0);
  if (!dl_get_symtab()) {
    __real_fprintf(MCOUT, "Cannot get Symbols (nm) !\n");
    __real_exit(-1);
    }
#if 0
  if (!get_symtab(mcglobal.progname, NULL)) { 
    __real_fprintf(MCOUT, "Cannot get Symbols from %s (nm) !\n", mcglobal.progname);
    __real_exit(-1);
    }

#ifdef __CYGWIN__
  __real_sprintf(buffer, "%s.exe", mcglobal.progname);
  if (!load_libs(buffer)) { 
#else
  if (!load_libs(mcglobal.progname)) { 
#endif
    __real_fprintf(MCOUT, "Cannot get library symbols from %s (ldd) !\n", mcglobal.progname);
    __real_exit(-1);
    }
#endif

  if (mcglobal.verbose) {
    __real_fprintf(MCOUT, "\nMemCheckInit(), verbose=%d, ignore=%d, mode=%x (hex)!\n", mcglobal.verbose, mcglobal.abort_ignore, mode);
    __real_fprintf(MCOUT, "\nMemCheckInit(), sizeof(bcmalloc)=%d, sizeof(BCMDUMMY)=%d !\n", sizeof(bcmalloc), sizeof(BCMDUMMY));
    if (mcglobal.checked_libs!=NULL) __real_fprintf(MCOUT, "\nMemCheckInit(), MCCHECKEDLIBS='%s'\n", mcglobal.checked_libs);
    }
  if (mcglobal.verbose>2) {
    __real_fprintf(MCOUT, "memory layout:\n");
    __real_fprintf(MCOUT, "%lx, main base pointer\n", mcglobal.main_bp);
    }

	// assert(mode==0);
  __real_fprintf(MCOUT, "MemCheckInit end ...\n");
  fflush(MCOUT);
  }

static void MemCheckInitArgv(int argc, char** argv, char **env) {
  int i;
  char *envptr;

  // mcset_progname(argv[0]);
  MemCheckInit();

  chkr_set_right(argv, sizeof(char*)*argc, MCRO);
  for (i=0; i<argc; i++) {
    chkr_set_right(argv[i], strlen(argv[i])+1, MCRO);
    }
  i=0;
  if (env!=NULL)
    while ((envptr=(char*)env[i])!=NULL) {
      chkr_set_right(envptr, strlen(envptr)+1, MCRW);
      i++;
      }
#ifdef WIN32
  // set rights for stdin, stdout, stderr
  chkr_set_right(&my_impure_ptr->_stdin, 4, MCRO);
  chkr_set_right(&my_impure_ptr->_stdout, 4, MCRO);
  chkr_set_right(&my_impure_ptr->_stderr, 4, MCRO);
  chkr_set_right(my_impure_ptr->_stdin, sizeof(FILE), MCRO);
  chkr_set_right(my_impure_ptr->_stdout, sizeof(FILE), MCRO);
  chkr_set_right(my_impure_ptr->_stderr, sizeof(FILE), MCRO);
#else
  #ifdef STDPOINTER
    chkr_set_right(&stdin, 4, MCRO);
    chkr_set_right(&stdout, 4, MCRO);
    chkr_set_right(&stderr, 4, MCRO);
  #endif
  chkr_set_right(stdin, sizeof(FILE), MCRO);
  chkr_set_right(stdout, sizeof(FILE), MCRO);
  chkr_set_right(stderr, sizeof(FILE), MCRO);
#endif

  // set a red zone around 'NULL', set first 1k of address space to UnkNown
  chkr_set_right(0, 1024, MCUN);
  }

#ifdef WIN32
typedef int (*main_ptr)(int, char**);
// our main, to be linked with crt0.o
int main(int argc, char** argv) {
  void *handele;
  main_ptr adr;
  int appresult;

  __real_fprintf(MCOUT, "MemCheck: argc: %d, argv[0]: %s\n", argc, argv[0]);

  // set our impure_ptr to global data:
  mcglobal.impure_ptr=my_impure_ptr;

  mcglobal.main_bp=getbp1();
  MemCheckInitArgv(argc, argv, NULL);

  __real_fprintf(MCOUT, "MemCheck::main() calling app main ...\n");
  handele=GetModuleHandle(NULL);
  if(handele==NULL) {
    __real_fprintf(MCOUT, "MemCheck: error on GetModuleHandle(NULL)\n");
    return -1;
    }
  if ((adr=(main_ptr)GetProcAddress(handele, "main"))==NULL) {
    __real_fprintf(MCOUT, "MemCheck: couldn't get symbol `main' from modulehandle\n");
    return -1;
    }
  __real_fprintf(MCOUT, "MemCheck: main - address: '%lx'\n", (ulong)adr);
  appresult=adr(argc, argv);
  MemCheckClose(appresult); // xxx // calls exit
  // xxx // not reached
  return appresult;
  }
#else

extern int __real_main(int, char**);
int __wrap_main(int argc, char **argv) {
  int appresult, wait_status=0;
  pid_t pid;

  __real_fprintf(MCOUT, "\nmain() called, argc=%d, real main is at: %lx\n", argc, (ulong)__real_main);
  mcglobal.main_bp=getbp1();
  MemCheckInitArgv(argc, argv, NULL);

  //ptrace pid=fork();
  //ptrace if (pid==0) {
    //ptrace fprintf(MCOUT, "ptrace cild - process ptracing main ...\n");
    //ptrace sleep(5);
    //ptrace if (ptrace(PTRACE_TRACEME)<0) {
    	//ptrace fprintf(MCOUT, "ptrace warning PTRACE_TRACEME failed\n");
	//ptrace };
    // appresult=main(argc, argv);
    appresult=__real_main(argc, argv);
    //ptrace argv[1]=NULL; // !!!
    //ptrace // fprintf(MCOUT, "execve %s ...\n", argv[0]);
    //ptrace sleep(5);
    //ptrace execve(argv[0], argv);
    //ptrace //  sleep(5);
    //ptrace //  fprintf(MCOUT, "ptrace ptraced fprintf ...\n");
    //ptrace // while (1) sleep(1);
    //ptrace fprintf(MCOUT, "ptrace return child - main\n");
    //ptrace sleep(10);
  //ptrace }
  //ptrace else {
    //ptrace // ptrace(PTRACE_ATTACH, pid, NULL, NULL);
    //ptrace // ptrace(PTRACE_SYSCALL, pid, 1, 0);
//ptrace // ptrace (PTRACE_SINGLESTEP, 0, 0, 0); 
    //ptrace while (1) {
      //ptrace fprintf(MCOUT, "ptrace parent - wait for child signal ...\n");
      //ptrace pid=wait(&wait_status);
      // pid=waitpid(pid, &wait_status, WNOHANG|WUNTRACED);
      //ptrace if (wait_status<0) {
        //ptrace fprintf(MCOUT, "ptrace WARNING parentwait error : %x\n", wait_status);
        //ptrace sleep(5);
      //ptrace }
      //ptrace if (WIFEXITED(wait_status)) {
      //ptrace fprintf(MCOUT, "ptrace child exit ...\n");
    //ptrace sleep(5);
        //ptrace MemCheckClose(appresult); // xxx // calls exit
        //ptrace break;
      //ptrace }
      //ptrace if (WIFSIGNALED(wait_status)) {
        //ptrace fprintf(MCOUT, "ptrace parent - receive signal ...\n");
    //ptrace sleep(5);
        //ptrace continue;
      //ptrace }
      //ptrace if (WIFSTOPPED(wait_status)) {
        //ptrace fprintf(MCOUT, "ptrace chiled is stopped...\n");
        //ptrace ptrace(PTRACE_SYSCALL, pid, 0, 0); 
        //ptrace continue;
      //ptrace }
      //ptrace fprintf(MCOUT, "ptrace WARNING unknown wait status: %x\n", wait_status);
      //ptrace sleep(5);
      //ptrace continue;
    //ptrace }
    //ptrace ptrace(PTRACE_DETACH, pid, NULL, NULL);
    //ptrace fprintf(MCOUT, "ptrace return parent - main: %x\n", wait_status);
  //ptrace }
  // xxx // not reached
  //ptrace sleep(5);
  return appresult;
  }
#endif

#ifdef WIN32
extern struct _reent *_imp__reent_data;
 
int WINAPI mcmain_dll_entry(HANDLE h, DWORD reason, void *ptr) {
  my_impure_ptr = _imp__reent_data;
  return 1;
  }

asm(".section .idata$3\n" ".long 0,0,0,0, 0,0,0,0");

struct _reent *mc_get_impure_ptr() {
  return my_impure_ptr;
  }
#endif

