/*
 * MemCheck, a memory-checker for C/C++.
 * 
 * This file is released under the terms of the GNU General Public License
 * as they appear in http://www.gnu.org/licenses/gpl.txt.
 * 
 * Copyright 2001, 2002 erwin.rieger@ibrieger.de
 */

//
// Trap calls to unchecked library functions, try to set memory rights in caller.
//
//
//
// We use a xxx to record all those functions, used to store additional information (
// func code, return address, status stuff).
//
// We use a stack of called ucl functions
//
// No threads !
//

#include <stdio.h>
#include <dlfcn.h>
#include <sys/mman.h>
#include <assert.h>
#include <errno.h>
#include <asm/param.h> // pagesize

#define OPCCALL 0xe8

// note: struct func_code is written over trapped functions, so 
// function size is important. Should be no prob since gcc seems 
//to align functions at 16 bit boundaries (fill bytes after ret insn).
#define UCLJUPMPOFS 5 // opcode + reladr
struct func_code {

	// unsigned char push1;
	// unsigned short mov1;

	unsigned char opcode;
	int	reladr;
	unsigned int func_info_ptr;	// used to pass address of func_info from
				// instrument_fun to trap
} __attribute__ ((packed));

struct frame_ovl {
	unsigned int prev_frame; // pushed bp of caller
	unsigned int ret; // return addr of us
} __attribute__ ((packed));

struct func_info {
	struct func_code *fun_code;	// the saved func code
	unsigned int fun_addr;		// start address
	unsigned int fun_ret;		// saved return address
	// unsigned int fun_client_bp;	// saved return client bp
	int check;			// status should function be checked
};

// stack of called ucl fcns, array of pointers to func_info's
// ucl_funcstack_len points to top
#define UCLFUNCSTACSIZE 0xffff  // 64 k pointers
static struct func_info *ucl_funcstack[UCLFUNCSTACSIZE];
static unsigned int ucl_funcstack_len=0;
// push func_info onto func stack:
static void ucl_pushfunc(struct func_info *f) {
	ucl_funcstack[ucl_funcstack_len++]=f;
	printf("ucl_pushfunc, len now: %d\n", ucl_funcstack_len);
}
// pop func_info from func stack, removes element from
// stack if remove is set:
static struct func_info *ucl_popfunc(int remove) {
	struct func_info *f=ucl_funcstack[ucl_funcstack_len-1];
	if (remove) ucl_funcstack_len--;
	return f;
}

static void ucl_endtrap();

// note: trapped func has no stack frame, situation on stack:
// - local space of client
// - return address of client
// - return address of trapped func !
// - the bp of the client that we have pushed
static void ucl_trap() {
	struct frame_ovl *ourframe;
	unsigned int ax, low_stack, up_stack, *ptr;
	struct func_info *fun_info;
	struct func_code *fun_code;

	__asm__ __volatile__ ("pushl %eax"); // needed ???
	// save ax, points to client stack
        __asm__ __volatile__("movl %%eax, %0": "=g" (ax));

	// get addr start of trapped function through our return address
	ourframe=__builtin_frame_address(0);
	// read func data:
	fun_code=(struct func_code*)(ourframe->ret-UCLJUPMPOFS);
	fun_info=(struct func_info*)fun_code->func_info_ptr;
	printf("trap func funcode: %x, func_info_ptr: %x\n", fun_code, fun_info);

	assert(fun_code->opcode==OPCCALL);
	assert(fun_info->fun_addr==(unsigned int)fun_code);
	assert(fun_info->fun_code->opcode==0x55);

	// restore start of trapped function:
	// printf("copy back: to %x from %x, %x\n", fun_code, fun_info->fun_code, ((char*)fun_info->fun_code)[0]);
	memcpy(fun_code, fun_info->fun_code, sizeof(struct func_code));

	// Modify our frame, restart trapped func:
	ourframe->ret-=5; // size of call
	// Store pointer, we use the space where the client has stored it's caller bp !
	// printf("ucl_trap, client bp at: %x\n", ourframe->prev_frame);
	// ptr=(unsigned int*)(ourframe->prev_frame);
	// fun_info->fun_client_bp=ptr[0]; // save 4 bytes
	// ptr[0]=(unsigned int)fun_info;

	// save return address and modify return address to call ucl_endtrap
	ptr=(unsigned int*)ourframe;
	fun_info->fun_ret=ptr[2];
	ptr[2]=(unsigned int)ucl_endtrap; // overwrite

	// update func stack
	ucl_pushfunc(fun_info);

// here is the memory testing:
        // lower boundary for ax test:
	low_stack=(unsigned int)ourframe + 12; // BP + ret + ret
        // upper boundary for ax test:
        up_stack=(unsigned int)ourframe->prev_frame;
	if ((ax>=low_stack) && (ax<up_stack)) {
		low_stack=ax; // !!
 		printf("client seems to use ax method, up: %x, low: %x, len: %d\n", up_stack, low_stack, up_stack-low_stack);
	}
	__asm__ __volatile__ ("popl %eax"); // needed ???
}

static void ucl_endtrap() {
	// struct frame_ovl *f;
	struct func_code *fun_code;
	struct func_info *fun_info;
	int *ptr;

	// __asm__ __volatile__ ("pushl %eax"); // needed ???
	// f=__builtin_frame_address(0);
	// get pointer to func info
	// printf("endtrap, client bp at: %x\n", f->prev_frame);
	// printf("endtrap, saved client bp: %x\n", fun_info->fun_client_bp);
	// ptr=(unsigned int*)f->prev_frame;

	// fun_info=(struct func_info*)ptr[0];
	// get fun_info, update func stack
	fun_info=ucl_popfunc(1);

	// restore bp of client
	// ptr[0]=fun_info->fun_client_bp;

	printf("endtrap called\n");
	fflush(stdout);
	if (fun_info->check) {
		// re - trap
		fun_code=(struct func_code*)fun_info->fun_addr;
		fun_code->opcode=OPCCALL;
		fun_code->reladr=(int)ucl_trap-((int)fun_code+UCLJUPMPOFS);
		fun_code->func_info_ptr=(unsigned int)fun_info;
		// free fun_info, fun_code, no longer needed, xxx ref in ucl_funcstack
	}

	// We are 'called' by a return, so we can't simpliy return here:
	// use a jump:
	// hoping, we can misuse ax
	// printf("endtrap: return to %x\n", fun_info->fun_ret);
        __asm__ __volatile__("movl %0, %%eax": : "g" (fun_info->fun_ret));
	// cleanup, jumped over
	__asm__ __volatile__ ("movl %ebp, %esp");
	__asm__ __volatile__ ("pop %ebp");
	__asm__ __volatile__ ("jmp *%eax");
}

void ucl_instrument_func(void *fun) {
	struct func_code *fun_code=fun, *saved_fun_code=(struct func_code*)malloc(sizeof(struct func_code));
	struct func_info *fun_info=(struct func_info*)malloc(sizeof(struct func_info));

	// note: no recursion prob, jump code is removed while trapped
	// func is executed
	assert(fun_code->opcode!=OPCCALL); // ups

	printf("instrumenting func at %x, sizeof func_code: %d\n", fun, sizeof(struct func_code));
	if (mprotect((void*)(((unsigned int)fun/EXEC_PAGESIZE)*EXEC_PAGESIZE), EXEC_PAGESIZE, PROT_WRITE)!=0) {
		printf("error on mprotect: %d, %s\n", errno, sys_errlist[errno]);
		exit(-1);
	}

	printf("instrumenting func fun_ptr: %x\n", fun_info);
	printf("instrumenting func saved code at: %x\n", saved_fun_code);

	// save startup code of trapped func
	memcpy(saved_fun_code, fun_code, sizeof(struct func_code));
	fun_info->fun_code=saved_fun_code;
	fun_info->fun_addr=(unsigned int)fun;
	fun_info->check=1;

	// overwrite beginning of func, the initialisation of stack frame
	fun_code->opcode=OPCCALL;
	// To compute the relative address to jump to, we need the address of the
	// target (which is function ucl_trap) and a jump start address. Jump start
	// is  current ip after reading call, that is: fun + 5 bytes jump code
	fun_code->reladr=(int)ucl_trap-((int)fun+UCLJUPMPOFS);
	fun_code->func_info_ptr=(unsigned int)fun_info;
	// printf("jumpaddress to trap func: %x, %x\n", &(j->opcode)+j->reladr, ucl_trap);
}


