/*
 * MemCheck, a memory-checker for C/C++.
 * 
 * This file is released under the terms of the GNU General Public License
 * as they appear in http://www.gnu.org/licenses/gpl.txt.
 * 
 * Copyright 2001, 2002 erwin.rieger@ibrieger.de
 */


#include <memcheck.h>
#include <time.h>
#include <stdarg.h>

extern struct mcglob * mcglobalP;

#define IsRet(ch) ((ch==0xc2)||(ch==0xc3)||(ch==0xca)||(ch==0xcb)||(ch==0xcf))

int __wrap_fprintf(FILE *f, const char *fmt, ...) {
  va_list param;
  va_start (param, fmt);
  return vfprintf(f, fmt, param);
}

void* __wrap_dlsym(void *handle, char *name) {
  void *res;
  mclib_list s;
  // int i;

  __real_fprintf(MCOUT, "MemCheck: dlsym(%lx, %s)\n", (ulong)handle, name);
  chkr_check_addr(&handle, sizeof(void *), MCRO);
  chkr_check_addr(&name, sizeof(char *), MCRO);
  MCStrlen(name, "__wrap_dlsym, parameter name", 0);

  // known handle ?
  s.handle=(ulong)handle;
  check_abort(list_find(&getMCGlobalPtr()->libs, &s)==NULL, "unknown handle in __wrap_dlsym");

  res=get_sym(handle, name, "__wrap_dlsym");
  return res;
  }

size_t __wrap_strftime(char *str, size_t max, const char *fmt, const struct tm *timp) {
  size_t res;

  chkr_check_addr(&str, sizeof(char *), MCRO);
  chkr_check_addr(&max, sizeof(size_t), MCRO);
  chkr_check_addr(&fmt, sizeof(char *), MCRO);
  chkr_check_addr(&timp, sizeof(struct tm *), MCRO);
  MCStrlen(fmt, "__wrap_strftime, parameter 3", 0);
  chkr_check_addr(timp, sizeof(struct tm), MCRO);

  res=__real_strftime(str, max, fmt, timp);
  if (res>0) {
    chkr_check_addr(str, res+1, MCWO);
    }
  return res;
  }

struct tm * __wrap_localtime(const time_t *t) {
  struct tm * res;

  chkr_check_addr(&t, sizeof(time_t *), MCRO);
  chkr_check_addr(t, sizeof(time_t), MCRO);

  res=__real_localtime(t);
  if (res!=NULL) {
    chkr_set_right(res, sizeof(struct tm), MCRO);
    }
  return res;
  }

char * __wrap_dlerror() {
  char * res;
  res=__real_dlerror();
  if (res!=NULL) {
    chkr_set_right(res, __real_strlen(res)+1, MCRO);
    }
  return res;
  }

extern void * __real_memmove(void *, const void *name, size_t);
void * __wrap_memmove(void *dst, const void *src, size_t s) {
  // test read from src:
  chkr_check_addr(src, s, MCRO);
  // test write to dst, set to RW:
  chkr_check_addr(dst, s, MCWO);
  
  return __real_memmove(dst, src, s);
  }

extern int __real_gethostname(char *, size_t);
int __wrap_gethostname(char *n, size_t s) {
  // test write to buffer:
  chkr_check_addr(n, s, MCWO);
  return __real_gethostname(n, s);
  }

char * __wrap_strdup(const char *s) {
  int i;
  char *new;

  i=MCStrlen(s, "__wrap_strdup, str", 0); /* args testen */
  new=MCMalloc(i+1, "__wrap_strdup", 0);
  MCStrcpy(new, s, "__wrap_strdup", 0);
  return new;
  }

extern size_t __real_read(int, void*, size_t);
size_t __wrap_read(int fd, void *DATA, size_t SIZE) {
  int i;

  i=__real_read(fd, DATA, SIZE);
  if (i<=0) return i;

  chkr_check_addr(DATA, i, MCWO);

  return i;
  }

extern char* __real_strstr(const char *, const char *);
char * __wrap_strstr(const char *s, const char *search) {
  MCStrlen(s, "__wrap_strstr, str", 0); /* args testen */
  MCStrlen(search, "__wrap_strstr, search", 0); /* args testen */
  return __real_strstr(s, search);
  }

extern double __real_modf(double x, double *iptr);
double __wrap_modf(double x, double *iptr) {
  chkr_check_addr(&x, sizeof(double), MCRO);
  chkr_check_addr(iptr, sizeof(double), MCWO);
  return __real_modf(x, iptr);
  }

#include <sys/time.h>
extern int __real_gettimeofday(struct timeval *, struct timezone *);
int __wrap_gettimeofday(struct timeval *t, struct timezone *z) {
 	chkr_check_addr(t, sizeof(struct timeval), MCWO);
	if (z!=NULL)
  	chkr_check_addr(z, sizeof(struct timezone), MCWO);
  return __real_gettimeofday(t, z);
  }

//
// c++:
//

// extern size_t __real__IO_sgetn (FILE *, void *, size_t);
// size_t __wrap__IO_sgetn(FILE *f, void *m, size_t s) {
	// int read;
	// chkr_check_addr(f, sizeof(FILE), MCRO);
	// read=__real__IO_sgetn(f, m, s);
	// chkr_set_right(m, read, MCRW);
	// return read;
// }



